//Version 2.1

#include "Fontain.h"

//This function is used to create the water pipe   
static SoSeparator* Pipe(float pos_x, float pos_y, float pos_z, float radius, float height)
{
  SoSeparator* sep = new SoSeparator();

  SoTexture2* texture = new SoTexture2();
  texture->filename.setValue("textures/metal.jpg");

  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(radius/2,height/2,radius/2);
 
  SoRotation* rot = new SoRotation();
  rot->rotation.setValue(SbVec3f(1.0f,0.0f,0.0f),(float)M_PI/16);
  
  SoTransform* transform = new SoTransform();
  transform->translation.setValue(pos_x,pos_y,pos_z);

  sep->addChild(texture);
  sep->addChild(transform);
  sep->addChild(rot);
  sep->addChild(scale);
  sep->addChild(new SoCylinder());

  return sep;
}

//This function is used to create the cylinder for the water jet
static SoSeparator* Cylinder(float pos_x, float pos_y, float pos_z, float radius, float height)
{
  SoSeparator* sep = new SoSeparator();

  SoTexture2* texture = new SoTexture2();
  texture->filename.setValue("textures/metal.jpg");

  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(radius/2,height/2,radius/2);

  SoRotation* rot = new SoRotation();
  rot->rotation.setValue(SbVec3f(1.0f,0.0f,0.0f),(float)M_PI/2);

  SoTransform* transform = new SoTransform();
  transform->translation.setValue(pos_x,pos_y,pos_z);

  
  sep->addChild(texture);
  sep->addChild(transform);
  sep->addChild(rot);
  sep->addChild(scale);
  sep->addChild(new SoCylinder());

  return sep;
}

//This function is used to create the boundaries
static SoSeparator* Cube(float pos_x, float pos_y, float pos_z, float length, float width, float height)
{

  SoSeparator* sep = new SoSeparator();
  
  SoTransform* transform = new SoTransform();
  transform->translation.setValue(pos_x,pos_y,pos_z);
    
  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(length/2,width/2,height/2);

  SoTexture2* texture = new SoTexture2();

  //DOESN'T WORK 
  //texture->wrapT.setValue(SoTexture2::Wrap::CLAMP);
  //texture->wrapS.setValue(SoTextureImageElement::REPEAT);
  texture->wrapS=SoTextureImageElement::REPEAT;
  //texture->model.setValue(SoTexture2::Model::REPLACE);
  texture->filename.setValue("textures/marble.jpg");

  sep->addChild(texture);
  sep->addChild(transform);
  sep->addChild(scale);
  sep->addChild(new SoCube());

  return sep;
}

//This function is used to create the bottom
static SoSeparator* Bottom(float pos_x, float pos_y, float pos_z, float length, float width, float height)
{
  SoSeparator* sep = new SoSeparator();
  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(length/2,width/2,height/2);

  SoTransform* transform = new SoTransform();
  transform->translation.setValue(pos_x,pos_y,pos_z);

  SoTexture2* texture2 = new SoTexture2();
  texture2->filename.setValue("textures/stone.jpg");

  sep->addChild(texture2);
  sep->addChild(transform);
  sep->addChild(scale);
  sep->addChild(new SoCube());

  return sep;
}



//This function is used to create the water
static SoSeparator* Water(WaveSurfaceKit * waveKit, float xScaling, float yScaling, float xTranslate, float yTranslate, float zTranslate)
{
  SoSeparator* sep = new SoSeparator();

  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(xScaling,yScaling,1);

  SoTransform* transform = new SoTransform();
  transform->translation.setValue(xTranslate,yTranslate,zTranslate);

  waveKit->startAnimation();

  sep->addChild(scale);
  sep->addChild(transform);
  sep->addChild(waveKit);

  return sep;
}


static void draw(){

  //First floor
  //The part standing
  root->addChild(Cube(0.0f,0.0f,3.5f,10.5f,0.5f,10.0f));
  //Front part
  root->addChild(Cube(0.0f,-10.0f,0.0f,10.5f,0.5f,3.0f));
  //Left part
  root->addChild(Cube(-5.0f,-5.0f,0.0f,.5f,9.5f,3.0f));
  //Right part
  root->addChild(Cube(5.0f,-5.0f,0.0f,.5f,9.5f,3.0f));
  //Bottom
  root->addChild(Bottom(0.0f,-5.0f,-1.45f,10.0f,10.0f,0.1f));
  root->addChild(Cube(0.0f,-5.0f,-1.625f,10.5f,10.5f,0.25f));
  //Water
  water1 = new WaveSurfaceKit(40, 5, 5, 0, -1,3);
  root->addChild(Water(water1,5,5,0,-1,0));

  //2nd floor
  //Front part
  root->addChild(Cube(0.0f,-7.0f,3.5f,6.5f,0.5f,2.0f));
  //Left part
  root->addChild(Cube(-3.0f,-3.5f,3.5f,0.5f,6.5f,2.0f));
  //Right part
  root->addChild(Cube(3.0f,-3.5f,3.5f,0.5f,6.5f,2.0f));
  //Bottom
  root->addChild(Bottom(0.0f,-3.5f,2.55f,6.0f,7.0f,0.1f));
  root->addChild(Cube(0.0f,-3.75f,2.375f,6.5f,7.0f,0.25f));
  //Water
  water2 = new WaveSurfaceKit(30, 3, 3.5, 0, -1,3);
  root->addChild(Water(water2,3,3.5,0,-1,3.5));
  //Water tube
  root->addChild(Pipe(0,-7.25,3.5,0.5,1.5));

  //3nd floor
  //Front part
  root->addChild(Cube(0.0f,-4.0f,6.5f,4.5f,0.5f,2.0f));
  //Left part
  root->addChild(Cube(-2.0f,-2.0f,6.5f,0.5f,3.5f,2.0f));
  //Right part
  root->addChild(Cube(2.0f,-2.0f,6.5f,0.5f,3.5f,2.0f));
  //Bottom
  root->addChild(Bottom(0.0f,-2.0f,5.55f,4.0f,4.0f,0.1f));
  root->addChild(Cube(0.0f,-2.25f,5.375f,4.5f,4.0f,0.25f));
  //Water
  water3 = new WaveSurfaceKit(20, 2, 2, 0, -1,3);
  root->addChild(Water(water3,2,2,0,-1,6.5));
  //Water tube
  root->addChild(Pipe(0,-4.25,6.5,0.5,1.5));

  // water jet's basement
  root->addChild(Cylinder(0.0f,-2.25f,5.975f,1.0f,0.75f));
  root->addChild(Cylinder(0.0f,-2.25f,6.6f,0.5f,0.5f));
  
  //Add the the sphere to the vector
  wf1.push_back(new WFElement(0.0f,-8.1f,3.35f,0.1f,0.25f,-0.0f,water1,water2));
  wf2.push_back(new WFElement(0.0f,-5.1f,6.35f,0.1f,0.25f,3.5f,water2,water3));
  wj.push_back(new WJElement(0.0f,-2.25f,6.6f,0.1f,2.2f,6.5f,water3));
  
  //Display the first water particles (for the 2 waterfall and the jet)
  root->addChild((wf1[0]->add2tree()));
  root->addChild((wf2[0]->add2tree()));
  root->addChild((wj[0]->add2tree()));
}

// handle keyboard events
static void event_cb(void * ud, SoEventCallback * eventCB){
  const SoEvent * event = eventCB->getEvent();
  if (SO_KEY_PRESS_EVENT(event, J)) {
	 if (enableJet==1){
		 enableJet=0;
	 }else{
		 enableJet=1;
	 }
	 eventCB->setHandled();
  }
}

//Timer CallBack: Modify the position of each particle(sphere) and add new ones.
static void TimerSensorCB(void*, SoSensor*){

	std::vector<WFElement*>::iterator it;
	std::vector<WFElement*>::iterator it2;
	std::vector<WJElement*>::iterator it3;
	//Move the particles
	//Lower Water Fall
	for( it = wf1.begin(); it != wf1.end(); it++ ) {
		(*it)->update(time_step);
	}
	//Upper Water Fall
	for( it2 = wf2.begin(); it2 != wf2.end(); it2++ ) {
		(*it2)->update(time_step);
	}
	//Jet
	for( it3 = wj.begin(); it3 != wj.end(); it3++ ) {
		(*it3)->update(time_step,enableJet);
	}

	//Add particles to water fall
	if (wf1.size()<54){
		wf1.push_back(new WFElement(random_val(0.0f,0.05f),random_val(-8.1f,0.05f),random_val(3.35f,0.05f),0.1f,random_val(0.4f,0.05f),0.0f,water1,water2));
		root->addChild(wf1[wf1.size()-1]->add2tree());
		wf1.push_back(new WFElement(random_val(0.0f,0.05f),random_val(-8.1f,0.05f),random_val(3.35f,0.05f),0.1f,random_val(0.4f,0.05f),0.0f,water1,water2));
		root->addChild(wf1[wf1.size()-1]->add2tree());
	}
	
	if (wf2.size()<50){
		wf2.push_back(new WFElement(random_val(0.0f,0.05f),random_val(-5.1f,0.05f),random_val(6.35f,0.05f),0.1f,random_val(0.4f,0.05f),3.5f,water2,water3));
		root->addChild(wf2[wf2.size()-1]->add2tree());
		wf2.push_back(new WFElement(random_val(0.0f,0.05f),random_val(-5.1f,0.05f),random_val(6.35f,0.05f),0.1f,random_val(0.4f,0.05f),3.5f,water2,water3));
		root->addChild(wf2[wf2.size()-1]->add2tree());
	}
	//Add particles to water jet
	if (wj.size()<150){
		wj.push_back(new WJElement(random_val(0.0f,0.05f),random_val(-2.25f,0.05f),6.6f,0.1f,random_val(2.2f,0.1f),6.5f,water3));
		root->addChild(wj[wj.size()-1]->add2tree());
		wj.push_back(new WJElement(random_val(0.0f,0.01f),random_val(-2.25f,0.01f),6.6f,0.1f,random_val(2.2f,0.1f),6.5f,water3));
		root->addChild(wj[wj.size()-1]->add2tree());
	}
}

int main(int, char ** argv){

  HWND window = SoWin::init(argv[0]);
  if (window==NULL) exit(1);

  enableJet=1;

  //coin_setenv("COIN_SHOW_FPS_COUNTER", "1", TRUE);

  SoWinExaminerViewer * viewer = new SoWinExaminerViewer(window);

  //Water
  WaveSurfaceKit::initClass();

  root = new SoSeparator;
  root->ref();

  //Ambient texture
  SoTexture2* AmbientTexture = new SoTexture2();
  AmbientTexture->filename.setValue("textures/stone.jpg");
  root->addChild(AmbientTexture);

  //Define the light (spot)
  SoSpotLight* l1= new SoSpotLight();
  l1->location.setValue(-10,-15,10);
  l1->direction.setValue(1,1,-1);
  l1->color.setValue(1.0,1.0,1.0);
  root->addChild(l1);

  //Define and enable the TimerSensor (called each Time interval)
  SoTimerSensor* simulation = new SoTimerSensor(TimerSensorCB, &wf1);
  simulation->setInterval(0.04);
  simulation->schedule();

  //Draw the scene
  draw();

  SoEventCallback * ecb = new SoEventCallback;
  ecb->addEventCallback(SoEvent::getClassTypeId(), event_cb, viewer);
  root->addChild(ecb);

  viewer->setSceneGraph(root);
  viewer->setFullScreen(TRUE);
  viewer->setDecoration(FALSE);

  viewer->show();
  //Necessary for the water to look like water
  viewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
  SoWin::show(window);
  SoWin::mainLoop();
  delete viewer;
  root->unref();
  return 0;
}
