// Version 2.1

#include <Inventor/Win/SoWin.h>
#include <Inventor/Win/viewers/SoWinExaminerViewer.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoMaterial.h>

#include "WFElement.h"

//Constructor
WFElement::WFElement(float pos_x, float pos_y, float pos_z, float radius, float speed, float end_z, WaveSurfaceKit * water_down, WaveSurfaceKit * water_up){
  this->pos_x=pos_x;
  this->pos_y=pos_y;
  this->pos_z=pos_z;
  this->radius=radius;
  init_x=pos_x;
  init_y=pos_y;
  init_z=pos_z;
  v_x=0;
  v_y=-speed;
  v_z=0;
  this->speed=speed;
  this->end_z=end_z;
  this->water_down=water_down;
  this->water_up=water_up;
}

//Add the sphere to the tree (coin3d)
SoSeparator* WFElement::add2tree(){
  SoSeparator* sep = new SoSeparator();

  SoMaterial* mat= new SoMaterial();
  mat->emissiveColor.setValue(0.9f,0.9f,0.9f);
  mat->specularColor.setValue(0.9f,0.9f,0.9f);
  mat->transparency=0.7f;

  SoScale* scale = new SoScale();
  scale->scaleFactor.setValue(radius,radius,radius);

  transform = new SoTransform();
  transform->translation.setValue(pos_x,pos_y,pos_z);

  sep->addChild(mat);
  sep->addChild(transform);
  sep->addChild(scale);
  sep->addChild(new SoSphere());
  return sep;
}

//Move the position of the sphere
void WFElement::move(float pos_x, float pos_y, float pos_z){
	this->pos_x=pos_x;
	this->pos_y=pos_y;
	this->pos_z=pos_z;
	transform->translation.setValue(pos_x,pos_y,pos_z);
}

//Update the position of the sphere using physics rules
void WFElement::update(float t){

	float val;

	pos_y+=v_y*t;
	pos_z+=v_z*t;

	v_z-=radius*10*t;

	//if the object reach the surface of the water,put it up back
	if (pos_z<end_z){
		pos_y=init_y,pos_z=init_z,v_z=0;
		// Why do we need this "-1"?
		water_down->reshape(pos_x,pos_y-1);
		val=water_up->getHeight();
		v_y=-speed-1.5f*val;
	}
	WFElement::move(pos_x,pos_y,pos_z);
}

